use std::io::{self, Read, Write};
use super::*;

#[derive(Debug)]
pub enum DecoderError {
    IO(io::Error),
    IllegalChar,
}

impl From<io::Error> for DecoderError {
    fn from(value: io::Error) -> Self {
        Self::IO(value)
    }
}

pub struct Decoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B32Alphabet,
}

impl<R: Read, W: Write> Decoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B32Alphabet>) -> Self {
        Self { reader, writer, alphabet: alphabet.unwrap_or_default() }
    }

    pub fn decode(&mut self) -> Result<(), DecoderError> {
        let mut buf = [0; 8];
        let mut num: u64 = 0;
        'outer: loop {
            if let Err(e) = self.reader.read_exact(&mut buf) {
                if e.kind() == io::ErrorKind::UnexpectedEof {
                    break;
                }
            }
            for c in &buf {
                let c = char::from(*c);
                num <<= 5;
                if !matches!(self.alphabet.pad(), Some(ch) if ch == c) {
                    let idx = self.alphabet.idx(c).ok_or(DecoderError::IllegalChar)?;
                    num |= idx as u64;
                }
            }
            let mut buf = [0; 5];
            for i in (0..5).rev() {
                let b = (num & 0xff) as u8;
                buf[i] = b;
                num >>= 8;
            }
            for c in &buf {
                if *c == b'\0' {
                    break 'outer;
                } else {
                    self.writer.write_all(&[*c])?;
                }
            }
        }
        self.writer.flush()?;
        Ok(())
    }
}

impl<R: Read> Decoder<R, Vec<u8>> {
    pub fn bytes(self) -> Vec<u8> {
        self.writer
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    static HELLO: &'static str = "Hello, World!";
    static ENCODED: &'static str = "JBSWY3DPFQQFO33SNRSCC===";

    #[test]
    fn get_idx() {
        let idx = B32_RFC4648_ALPHABET.idx('S').unwrap();
        assert_eq!(idx, 18);
    }

    #[test]
    fn hello() {
        let reader = ENCODED.as_bytes();
        let writer = Vec::<u8>::new();
        let mut decoder = Decoder::new(reader, writer, None);
        decoder.decode().unwrap();
        assert_eq!(HELLO, String::from_utf8(decoder.bytes()).unwrap());
    }
}

